// Copyright (C) 2009-2009 Martin Sandve Alnæs
// Licensed under the GNU LGPL Version 3.0.
//
// First added:  2009-01-01
// Last changed: 2009-04-01
//
// This demo program solves a hyperelasticity problem
// in 3D using the Saint-Vernant-Kirchoff model.

#include <iostream>
#include <vector>
#include <dolfin.h>
#include "generated_code/HyperElasticitySVK.h"

using dolfin::message;
using dolfin::Function;
using dolfin::MeshFunction;
using dolfin::FunctionSpace;
using dolfin::SubDomain;
using dolfin::uint;
using dolfin::UnitCube;
using dolfin::DirichletBC;
using dolfin::BoundaryCondition;
using dolfin::VariationalProblem;
using dolfin::Vector;
using dolfin::Matrix;
using dolfin::File;

using namespace HyperElasticitySVK;

using std::cout;
using std::endl;

// ---------------------------------------- Constant parameters

const double eps = DOLFIN_EPS;
const double a = 0.03;

// ---------------------------------------- Functions

class Value: public Function
{
public:
  Value(const FunctionSpace & V, double value):
    Function(V), value(value)
  {}

  void eval(double* values, const double* x) const
  {
    values[0] = value;
  }

  double value;
};

class BodyForce: public Function
{
public:
  BodyForce(const FunctionSpace & V): Function(V) {}

  void eval(double* values, const double* x) const
  {
    const double dx = x[0] - 0.5;
    const double dy = x[1] - 0.5;
    const double dz = x[2] - 0.5;

    const double lamda = 1.0;
    const double mu = 1.0;

//  values[0] =  sin(x[1])*mu*a+a*( sin(x[1])-x[1])+lamda*( cos(x[0])-1.0)*sin(x[0])*(a*a)+lamda*( cos(x[0])-1.0)*sin(x[2])*( cos(x[2])-1.0)*(a*a*a)+2.0*( cos(x[0])-1.0)*sin(x[0])*mu*(a*a);
//  values[1] =  lamda*( cos(x[0])-1.0)*sin(x[0])*(a*a*a)*( cos(x[1])-1.0)+2.0*sin(x[1])*mu*(a*a)*( cos(x[1])-1.0)+mu*sin(x[2])*a+( sin(x[2])-x[2])*a+lamda*sin(x[1])*(a*a)*( cos(x[1])-1.0);
//  values[2] = -( x[0]-sin(x[0]))*a+sin(x[0])*mu*a+2.0*mu*sin(x[2])*( cos(x[2])-1.0)*(a*a)+lamda*sin(x[2])*( cos(x[2])-1.0)*(a*a)+lamda*sin(x[1])*( cos(x[2])-1.0)*(a*a*a)*( cos(x[1])-1.0);

//  values[0] = x[0]*a;
//  values[1] = 0.0000000000000000e+00;
//  values[2] = 0.0000000000000000e+00;

//  values[0] = -lamda*( x[1]*a*x[2]+1.0)*( x[0]*(x[1]*x[1])*(a*a)+x[0]*(a*a)*(x[2]*x[2]))+x[0]*x[1]*a*x[2]-x[0]*mu*(x[1]*x[1])*(a*a)+-2.0*x[0]*( x[1]*a*x[2]+1.0)*mu*(x[1]*x[1])*(a*a)-x[0]*mu*(a*a)*(x[2]*x[2])+-2.0*x[0]*( x[1]*a*x[2]+1.0)*mu*(a*a)*(x[2]*x[2]);
//  values[1] = -a*( 2.0*(x[0]*x[0])*mu*(a*a)*(x[2]*x[2])+lamda*( (x[0]*x[0])*(a*a)*(x[2]*x[2])+(x[0]*x[0])*(x[1]*x[1])*(a*a)+pow( x[1]*a*x[2]+1.0,2.0)-1.0))*x[2]/2.0+-3.0*(x[0]*x[0])*mu*(x[1]*x[1])*(a*a*a)*x[2]-pow( x[1]*a*x[2]+1.0,2.0)*mu*a*x[2]-lamda*( ( x[1]*a*x[2]+1.0)*a*x[2]+(x[0]*x[0])*x[1]*(a*a))-x[0]*( lamda*( x[0]*(x[1]*x[1])*(a*a)+x[0]*(a*a)*(x[2]*x[2]))+2.0*x[0]*mu*(a*a)*(x[2]*x[2]))*a*x[2]-(x[0]*x[0])*mu*x[1]*(a*a);
//  values[2] = -pow( x[1]*a*x[2]+1.0,2.0)*mu*x[1]*a-(x[0]*x[0])*mu*(a*a)*x[2]-x[1]*a*( 2.0*(x[0]*x[0])*mu*(x[1]*x[1])*(a*a)+lamda*( (x[0]*x[0])*(a*a)*(x[2]*x[2])+(x[0]*x[0])*(x[1]*x[1])*(a*a)+pow( x[1]*a*x[2]+1.0,2.0)-1.0))/2.0-lamda*( (x[0]*x[0])*(a*a)*x[2]+( x[1]*a*x[2]+1.0)*x[1]*a)+-3.0*(x[0]*x[0])*mu*x[1]*(a*a*a)*(x[2]*x[2])-( 2.0*x[0]*mu*(x[1]*x[1])*(a*a)+lamda*( x[0]*(x[1]*x[1])*(a*a)+x[0]*(a*a)*(x[2]*x[2])))*x[0]*x[1]*a;

    values[0] = x[0]*a;
    values[1] = x[1]*a;
    values[2] = a*x[2];
  }
};

class BoundaryPressure: public Function
{
public:
  BoundaryPressure(const FunctionSpace & V): Function(V) {}

  void eval(double* values, const double* x) const
  {
    // TODO: Set some nonzero pressure
    values[0] = 0.0;
  }
};

class Displacement: public Function
{
public:
  Displacement(const FunctionSpace & V): Function(V) {}

  void eval(double* values, const double* x) const
  {
    const double dx = x[0] - 0.5;
    const double dy = x[1] - 0.5;
    const double dz = x[2] - 0.5;

    // Working:
//  values[0] =  0.05*dx;
//  values[1] = -0.05*dy;
//  values[2] =  0.0;

    // Not working:
//  values[0] =  0.1*dx;
//  values[1] = -0.1*dy;
//  values[2] =  0.0;

    // Not working:
//  values[0] =  sin(x[1]);
//  values[1] =  sin(x[0]);
//  values[2] =  0.0;

    // Not working:
//  values[0] = 0.1 * sin(x[1]);
//  values[1] = 0.1 * sin(x[0]);
//  values[2] = 0.1 * 0.0;

    // Working:
//  values[0] = 0.01 * sin(x[1]);
//  values[1] = 0.01 * sin(x[0]);
//  values[2] = 0.01 * 0.0;

//  values[0] = a*( sin(x[1])-x[1]);
//  values[1] = ( sin(x[2])-x[2])*a;
//  values[2] = -( x[0]-sin(x[0]))*a;

//  values[0] = x[0]*a;
//  values[1] = 0.0000000000000000e+00;
//  values[2] = 0.0000000000000000e+00;

//  values[0] = x[0]*x[1]*a*x[2];
//  values[1] = 0.0000000000000000e+00;
//  values[2] = 0.0000000000000000e+00;

    values[0] = x[0]*a;
    values[1] = x[1]*a;
    values[2] = a*x[2];
  }
};

// ---------------------------------------- Subdomains

class PressureBoundary: public SubDomain
{
  bool inside(const double* x, bool on_boundary) const
  {
    //return (x[0] < eps) || (x[0] > 1.0-eps) || (x[1] < eps) || (x[1] > 1.0-eps) || (x[2] < eps) || (x[2] > 1.0-eps);
    return false;
  }
};

class DirichletBoundary: public SubDomain
{
  bool inside(const double* x, bool on_boundary) const
  {
    //return x[0] < DOLFIN_EPS || x[0] > 1.0-DOLFIN_EPS;
    return (x[0] < eps) || (x[0] > 1.0-eps) || (x[1] < eps) || (x[1] > 1.0-eps) || (x[2] < eps) || (x[2] > 1.0-eps);
  }
};

// ---------------------------------------- Main program

int main(int argc, char**argv)
{
    // Geometry
    message("Mesh");
    unsigned n = 12;
    UnitCube mesh(n, n, n);

    // Boundaries
    message("Boundaries");
    unsigned pressure_boundary_marker  = 0;
    unsigned dirichlet_boundary_marker = 1;

    MeshFunction<unsigned> boundaries(mesh, mesh.geometry().dim()-1);

    PressureBoundary pressure_boundary;
    pressure_boundary.mark(boundaries, pressure_boundary_marker);
    
    DirichletBoundary dirichlet_boundary;
    dirichlet_boundary.mark(boundaries, dirichlet_boundary_marker);

    // Function spaces
    message("Function spaces");
    Form_a_F::TestSpace V(mesh);
    CoefficientSpace_sigma_0 P(mesh);
    CoefficientSpace_mu C(mesh);
    
    // Coefficient functions
    message("Functions");
    
    // Displacement at current and two previous timesteps
    Function u(V);
    Function up(V);
    Function upp(V);
    u.vector().zero();
    up.vector().zero();
    upp.vector().zero();

    // Time parameters
    Value dt(C, 1.0);
    
    // External forces
    BodyForce g(V);
    BoundaryPressure sigma_0(P);
    
    // Material parameters
    Value rho(C, 1.0);
    Value mu(C, 1.0);
    Value lamda(C, 1.0);
    
    // Analytical manufactured solution
    Displacement usol(V);

    // Collect coefficients
    CoefficientSet coeffs;
    coeffs.u       = u;
    coeffs.up      = up;
    coeffs.upp     = upp;
    coeffs.dt      = dt;
    coeffs.g       = g;
    coeffs.sigma_0 = sigma_0;
    coeffs.rho     = rho;
    coeffs.lamda   = lamda;
    coeffs.mu      = mu;

    coeffs.disp();    

    // Forms
    message("Forms");
    Form_a_f f(coeffs);
    Form_a_F L(V, coeffs);
    Form_a_J a(V, V, coeffs);

    // Setup boundary conditions
    message("Boundary conditions");
    
    // Dirichlet boundary conditions
    DirichletBC bc(V, usol, boundaries, dirichlet_boundary_marker);
    
    std::vector<BoundaryCondition*> bcs;
    bcs.push_back(&bc);
    
    // Create variational problem (a, L, bcs, cell_domains, exterior_facet_domains, interior_facet_domains, nonlinear)
    message("Variational problem");
    VariationalProblem problem(a, L, bcs, 0, &boundaries, 0, true);
    
    // Compute and show linear system
    if(false)
    {
        message("================================================================================");
        message("Printing linear system");
        Matrix J;
        problem.J(J, u.vector());
        J.disp();
        Vector F;
        problem.F(F, u.vector());
        F.disp();
        message("================================================================================");
    }

    // Solve!
    message("Solving");
    problem.set("linear solver", "direct");
    //problem.set("linear solver", "iterative");
    problem.solve(u);
    
    return 0;

    // Postprocessing
    message("Postprocessing");
    
    // Interpolate usol
    Function usol_h(V);
    usol.interpolate(usol_h.vector(), V);

    // Compute functional value of u and exact solution
    f.u = u;
    double f_value = assemble(f);
    f.u = usol_h;
    double f_value2 = assemble(f);

    // Print limits of u
    cout << endl;
    cout << "==========================" << endl;
    cout << "Norm of u_h = " << u.vector().norm() << endl;
    cout << "Min  of u_h = " << u.vector().min() << endl;
    cout << "Max  of u_h = " << u.vector().max() << endl;
    cout << endl;
    cout << "Norm of u_e = " << usol_h.vector().norm() << endl;
    cout << "Min  of u_e = " << usol_h.vector().min() << endl;
    cout << "Max  of u_e = " << usol_h.vector().max() << endl;
    cout << endl;
    cout << "f(u) = " << f_value << endl;
    cout << "f(u_exact) = " << f_value2 << endl;
    cout << "(u_exact is not correct)" << endl;
    cout << "==========================" << endl;

    // Compute F at u for verification
    Vector F;
    problem.F(F, u.vector());
    cout << endl;
    cout << "==========================" << endl;
    cout << "Norm of F = " << F.norm() << endl;
    cout << "Min  of F = " << F.min() << endl;
    cout << "Max  of F = " << F.max() << endl;
    cout << "==========================" << endl;
    
    // Compute e_h = usol_h - u
    Function e_h(V);
    e_h.vector() = u.vector();
    e_h.vector() -= usol_h.vector();
    cout << endl;
    cout << "==========================" << endl;
    cout << "Norm of e_h = " << e_h.vector().norm() << endl;
    cout << "Min  of e_h = " << e_h.vector().min() << endl;
    cout << "Max  of e_h = " << e_h.vector().max() << endl;
    cout << "==========================" << endl;
    
    // Forms for postprocessing
    
    // Project div(u)
    {
      Form_projection_a_div_u::TrialSpace V(mesh);
      Form_projection_a_div_u a(V, V, coeffs);
      Form_projection_L_div_u L(V, coeffs);
      VariationalProblem problem(a, L);
      problem.set("linear solver", "direct");

      Function u(V);
      problem.solve(u);
      File tmp_file("div_u.pvd");
      tmp_file << u;
    }

    // Project J = det F
    {
      Form_projection_a_J::TrialSpace V(mesh);
      Form_projection_a_J a(V, V, coeffs);
      Form_projection_L_J L(V, coeffs);
      VariationalProblem problem(a, L);
      problem.set("linear solver", "direct");

      Function u(V);
      problem.solve(u);
      File tmp_file("J.pvd");
      tmp_file << u;
    }

    // Project psi
    {
      Form_projection_a_psi::TrialSpace V(mesh);
      Form_projection_a_psi a(V, V, coeffs);
      Form_projection_L_psi L(V, coeffs);
      VariationalProblem problem(a, L);
      problem.set("linear solver", "direct");

      Function u(V);
      problem.solve(u);
      File tmp_file("psi.pvd");
      tmp_file << u;
    }

    //plot(w);
    
    // Write functions to file
    message("Writing to file");
    
    File ufile("u.pvd");
    ufile << u;
    
    File usolfile("usol.pvd");
    usolfile << usol_h;
    
    File efile("e.pvd");
    efile << e_h;

    message("Done!");
}

